# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from .base_wake_turbulence import WakeTurbulence


class CrespoHernandez(WakeTurbulence):
    """
    CrespoHernandez is a wake-turbulence model that is used to compute
    additional variability introduced to the flow field by operation of a wind
    turbine. Implementation of the model follows the original formulation and
    limitations outlined in :cite:`cht-crespo1996turbulence`.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: cht-
    """

    default_parameters = {
        "initial": 0.1,
        "constant": 0.5,
        "ai": 0.8,
        "downstream": -0.32,
    }

    def __init__(self, parameter_dictionary):
        """
        Stores model parameters for use by methods.

        Args:
            parameter_dictionary (dict): Model-specific parameters.
                Default values are used when a parameter is not included
                in `parameter_dictionary`. Possible key-value pairs include:

                -   **initial** (*float*): The initial ambient turbulence
                    intensity, expressed as a decimal fraction.
                -   **constant** (*float*): The constant used to scale the
                    wake-added turbulence intensity.
                -   **ai** (*float*): The axial induction factor exponent used
                    in in the calculation of wake-added turbulence.
                -   **downstream** (*float*): The exponent applied to the
                    distance downstream of an upstream turbine normalized by
                    the rotor diameter used in the calculation of wake-added
                    turbulence.
        """
        super().__init__(parameter_dictionary)
        self.model_string = "crespo_hernandez"
        model_dictionary = self._get_model_dict(__class__.default_parameters)

        # turbulence parameters
        self.ti_initial = model_dictionary["initial"]
        self.ti_constant = model_dictionary["constant"]
        self.ti_ai = model_dictionary["ai"]
        self.ti_downstream = model_dictionary["downstream"]

    def function(self, ambient_TI, coord_ti, turbine_coord, turbine):
        """
        Calculates wake-added turbulence as a function of
        external conditions and wind turbine operation. This function is
        accessible through the :py:class:`~.wake.Wake` class as the
        :py:meth:`~.Wake.turbulence_function` method.

        Args:
            ambient_TI (float): TI of the background flow field.
            coord_ti (:py:class:`~.utilities.Vec3`): Coordinate where TI
                is to be calculated (e.g. downstream wind turbines).
            turbine_coord (:py:class:`~.utilities.Vec3`): Coordinate of
                the wind turbine adding turbulence to the flow.
            turbine (:py:class:`~.turbine.Turbine`): Wind turbine
                adding turbulence to the flow.

        Returns:
            float: Wake-added turbulence from the current
                wind turbine (**turbine**) at location specified
                by (**coord_ti**).
        """
        ti_initial = ambient_TI

        # turbulence intensity calculation based on Crespo et. al.
        ti_calculation = (
            self.ti_constant
            * turbine.aI ** self.ti_ai
            * ti_initial ** self.ti_initial
            * ((coord_ti.x1 - turbine_coord.x1) / turbine.rotor_diameter)
            ** self.ti_downstream
        )

        # Update turbulence intensity of downstream turbines
        return ti_calculation

    @property
    def ti_initial(self):
        """
        Parameter that is the initial ambient turbulence intensity, expressed as
        a decimal (e.g. 10% TI -> 0.10).

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            ti_initial (float): Initial ambient turbulence intensity.

        Returns:
            float: Initial ambient turbulence intensity.

        Raises:
            ValueError: Invalid value.
        """
        return self._ti_initial

    @ti_initial.setter
    def ti_initial(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for " + "initial: {}, expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._ti_initial = value
        if value != __class__.default_parameters["initial"]:
            self.logger.info(
                (
                    "Current value of initial, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["initial"])
            )

    @property
    def ti_constant(self):
        """
        Constant parameter used to scale the wake-added turbulence
        intensity.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            ti_constant (float): Scales the wake-added turbulence intensity.

        Returns:
            float: Scales the wake-added turbulence intensity.

        Raises:
            ValueError: Invalid value.
        """
        return self._ti_constant

    @ti_constant.setter
    def ti_constant(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for " + "constant: {}, expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._ti_constant = value
        if value != __class__.default_parameters["constant"]:
            self.logger.info(
                (
                    "Current value of constant, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["constant"])
            )

    @property
    def ti_ai(self):
        """
        Axial induction factor exponent used in in the calculation of
        wake-added turbulence.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            ti_ai (float): Axial induction factor exponent for wake-added
              turbulence.

        Returns:
            float: Axial induction factor exponent for wake-added turbulence.

        Raises:
            ValueError: Invalid value.
        """
        return self._ti_ai

    @ti_ai.setter
    def ti_ai(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for " + "ai: {}, expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._ti_ai = value
        if value != __class__.default_parameters["ai"]:
            self.logger.info(
                (
                    "Current value of ai, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["ai"])
            )

    @property
    def ti_downstream(self):
        """
        Exponent applied to the distance from an upstream turbine normalized by
        the rotor diameter. Used in the calculation of wake-added turbulence.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            ti_downstream (float): Downstream distance exponent for
              wake-added turbulence.

        Returns:
            float: Downstream distance exponent for wake-added turbulence.

        Raises:
            ValueError: Invalid value.
        """
        return self._ti_downstream

    @ti_downstream.setter
    def ti_downstream(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for " + "downstream: {}, expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._ti_downstream = value
        if value != __class__.default_parameters["downstream"]:
            self.logger.info(
                (
                    "Current value of downstream, {0}, is not equal to "
                    + "tuned value of {1}."
                ).format(value, __class__.default_parameters["downstream"])
            )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import numpy as np

from ...utilities import cosd
from .base_velocity_deficit import VelocityDeficit


class MultiZone(VelocityDeficit):
    """
    The MultiZone model computes the wake velocity deficit based
    on the original multi-zone FLORIS model. See
    :cite:`mvm-gebraad2014data,mvm-gebraad2016wind` for more details.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: mvm-
    """

    default_parameters = {
        "me": [-0.5, 0.3, 1.0],
        "we": 0.05,
        "aU": 12.0,
        "bU": 1.3,
        "mU": [0.5, 1.0, 5.5],
    }

    def __init__(self, parameter_dictionary):
        """
        Stores model parameters for use by methods.

        Args:
            parameter_dictionary (dict): Model-specific parameters.
                Default values are used when a parameter is not included
                in `parameter_dictionary`. Possible key-value pairs include:

                -   **me** (*list*): A list of three floats that help determine
                    the slope of the diameters of the three wake zones
                    (near wake, far wake, mixing zone) as a function of
                    downstream distance.
                -   **we** (*float*): Scaling parameter used to adjust the wake
                    expansion, helping to determine the slope of the diameters
                    of the three wake zones as a function of downstream
                    distance, as well as the recovery of the velocity deficits
                    in the wake as a function of downstream distance.
                -   **aU** (*float*): A float that is a parameter used to
                    determine the dependence of the wake velocity deficit decay
                    rate on the rotor yaw angle.
                -   **bU** (*float*): A float that is another parameter used to
                    determine the dependence of the wake velocity deficit decay
                    rate on the rotor yaw angle.
                -   **mU** (*list*): A list of three floats that are parameters
                    used to determine the dependence of the wake velocity
                    deficit decay rate for each of the three wake zones on the
                    rotor yaw angle.
        """
        super().__init__(parameter_dictionary)
        self.model_string = "multizone"
        model_dictionary = self._get_model_dict(__class__.default_parameters)
        self.me = [n for n in model_dictionary["me"]]
        self.we = model_dictionary["we"]
        self.aU = model_dictionary["aU"]
        self.bU = model_dictionary["bU"]
        self.mU = [n for n in model_dictionary["mU"]]

    def function(
        self,
        x_locations,
        y_locations,
        z_locations,
        turbine,
        turbine_coord,
        deflection_field,
        flow_field,
    ):
        """
        Using the original FLORIS multi-zone wake model, this method
        calculates and returns the wake velocity deficits, caused by
        the specified turbine, relative to the freestream velocities at
        the grid of points comprising the wind farm flow field.

        Args:
            x_locations (np.array): An array of floats that contains the
                streamwise direction grid coordinates of the flow field
                domain (m).
            y_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the direction normal to
                x and parallel to the ground (m).
            z_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the vertical
                direction (m).
            turbine (:py:obj:`floris.simulation.turbine`): Object that
                represents the turbine creating the wake.
            turbine_coord (:py:obj:`floris.utilities.Vec3`): Object containing
                the coordinate of the turbine creating the wake (m).
            deflection_field (np.array): An array of floats that contains the
                amount of wake deflection in meters in the y direction at each
                grid point of the flow field.
            flow_field (:py:class:`floris.simulation.flow_field`): Object
                containing the flow field information for the wind farm.

        Returns:
            np.array, np.array, np.array:
                Three arrays of floats that contain the wake velocity
                deficit in m/s created by the turbine relative to the freestream
                velocities for the U, V, and W components, aligned with the
                x, y, and z directions, respectively. The three arrays contain
                the velocity deficits at each grid point in the flow field.
        """

        mu = self.mU / cosd(self.aU + self.bU * turbine.yaw_angle)

        # distance from wake centerline
        rY = abs(y_locations - (turbine_coord.x2 + deflection_field))
        # rZ = abs(z_locations - (turbine.hub_height))
        dx = x_locations - turbine_coord.x1

        # wake zone diameters
        nearwake = turbine.rotor_radius + self.we * self.me[0] * dx
        farwake = turbine.rotor_radius + self.we * self.me[1] * dx
        mixing = turbine.rotor_radius + self.we * self.me[2] * dx

        # initialize the wake field
        c = np.zeros(x_locations.shape)

        # near wake zone
        mask = rY <= nearwake
        c += (
            mask
            * (
                turbine.rotor_diameter
                / (turbine.rotor_diameter + 2 * self.we * mu[0] * dx)
            )
            ** 2
        )
        # mask = rZ <= nearwake
        # c += mask * (radius / (radius + we * mu[0] * dx))**2

        # far wake zone
        # ^ is XOR, x^y:
        #   Each bit of the output is the same as the corresponding bit in x
        #   if that bit in y is 0, and it's the complement of the bit in x
        #   if that bit in y is 1.
        # The resulting mask is all the points in far wake zone that are not
        # in the near wake zone
        mask = (rY <= farwake) ^ (rY <= nearwake)
        c += (
            mask
            * (
                turbine.rotor_diameter
                / (turbine.rotor_diameter + 2 * self.we * mu[1] * dx)
            )
            ** 2
        )
        # mask = (rZ <= farwake) ^ (rZ <= nearwake)
        # c += mask * (radius / (radius + we * mu[1] * dx))**2

        # mixing zone
        # | is OR, x|y:
        #   Each bit of the output is 0 if the corresponding bit of x AND
        #   of y is 0, otherwise it's 1.
        # The resulting mask is all the points in mixing zone that are not
        # in the far wake zone and not in  near wake zone
        mask = (rY <= mixing) ^ ((rY <= farwake) | (rY <= nearwake))
        c += (
            mask
            * (
                turbine.rotor_diameter
                / (turbine.rotor_diameter + 2 * self.we * mu[2] * dx)
            )
            ** 2
        )
        # mask = (rZ <= mixing) ^ ((rZ <= farwake) | (rZ <= nearwake))
        # c += mask * (radius / (radius + we * mu[2] * dx))**2

        # filter points upstream
        c[x_locations - turbine_coord.x1 < 0] = 0

        return (
            2 * turbine.aI * c * flow_field.wind_map.grid_wind_speed,
            np.zeros(np.shape(c)),
            np.zeros(np.shape(c)),
        )

    @property
    def me(self):
        """
        A list of three floats that help determine the slope of the diameters
        of the three wake zones (near wake, far wake, mixing zone) as a
        function of downstream distance.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (list): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._me

    @me.setter
    def me(self, value):
        # if type(value) is list and len(value) == 3 and \
        #                     all(type(val) is float for val in value) is True:
        #     self._me = value
        # elif type(value) is list and len(value) == 3 and \
        #                     all(type(val) is float for val in value) is False:
        #     self._me = [float(val) for val in value]
        # else:
        #     raise ValueError("Invalid value given for me: {}".format(value))

        if (
            type(value) is not list
            or len(value) != 3
            or not all(type(val) is float for val in value)
        ):
            err_msg = (
                "Invalid value type given for me: {}, " + "expected list of length 3."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._me = value
        if value != __class__.default_parameters["me"]:
            self.logger.info(
                (
                    "Current value of me, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["me"])
            )

    @property
    def we(self):
        """
        Scaling parameter used to adjust the wake expansion, helping to
        determine the slope of the diameters of the three wake zones as a
        function of downstream distance, as well as the recovery of the
        velocity deficits in the wake as a function of downstream distance.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._we

    @we.setter
    def we(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for we: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._we = value
        if value != __class__.default_parameters["we"]:
            self.logger.info(
                (
                    "Current value of we, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["we"])
            )

    @property
    def aU(self):
        """
        Parameter used to determine the dependence of the wake velocity deficit
        decay rate on the rotor yaw angle.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._aU

    @aU.setter
    def aU(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for aU: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._aU = value
        if value != __class__.default_parameters["aU"]:
            self.logger.info(
                (
                    "Current value of aU, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["aU"])
            )

    @property
    def bU(self):
        """
        Parameter used to determine the dependence of the wake velocity deficit
        decay rate on the rotor yaw angle.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._bU

    @bU.setter
    def bU(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for bU: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._bU = value
        if value != __class__.default_parameters["bU"]:
            self.logger.info(
                (
                    "Current value of bU, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["bU"])
            )

    @property
    def mU(self):
        """
        A list of three floats that are parameters used to determine the
        dependence of the wake velocity deficit decay rate for each of the
        three wake zones on the rotor yaw angle.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (list): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._mU

    @mU.setter
    def mU(self, value):
        # if type(value) is list and len(value) == 3 and \
        #                     all(type(val) is float for val in value) is True:
        #     self._mU = value
        # elif type(value) is list and len(value) == 3 and \
        #                     all(type(val) is float for val in value) is False:
        #     self._mU = [float(val) for val in value]
        # else:
        #     raise ValueError("Invalid value given for mU: {}".format(value))

        if (
            type(value) is not list
            or len(value) != 3
            or not all(type(val) is float for val in value)
        ):
            err_msg = (
                "Invalid value type given for mU: {}, " + "expected list of length 3."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._mU = value
        if value != __class__.default_parameters["mU"]:
            self.logger.info(
                (
                    "Current value of mU, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["mU"])
            )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

"""
The :py:obj:`floris.simulation` package contains the modules used to
complete a wake simulation for a given wind farm and turbine
configuration.

All classes can be be imported with

    >>> import floris.simulation

The ``__init__.py`` file enables the import of all modules in this
package so any additional modules should be included there.

Examples:
    >>> import floris.simulation

    >>> dir(floris.simulation)
    ['Farm', 'Floris', 'FlowField', 'InputReader', 'Turbine',
    'TurbineMap', 'Wake', 'WindMap', '__builtins__', '__cached__',
    '__doc__', '__file__', '__loader__', '__name__', '__package__',
    '__path__', '__spec__', 'farm', 'floris', 'flow_field', 'input_reader',
    'turbine', 'turbine_map', 'wake', 'wake_combination', 'wake_deflection',
    'wake_turbulence', 'wake_velocity', 'wind_map']
"""

from . import wake_velocity, wake_deflection, wake_turbulence, wake_combination
from .farm import Farm
from .wake import Wake
from .floris import Floris
from .turbine import Turbine
from .wind_map import WindMap
from .flow_field import FlowField
from .turbine_map import TurbineMap
from .input_reader import InputReader



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import numpy as np
from scipy.special import gamma

from ....utilities import cosd, sind, tand
from .gaussian_model_base import GaussianModel
from ..base_velocity_deficit import VelocityDeficit


class Blondel(GaussianModel):
    """
    Blondel is a direct implementation of the super-Gaussian model
    described in :cite:`bcv-blondel2020alternative` with GCH disabled by
    default. See :cite:`bcv-King2019Controls` for info on GCH.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: bcv-
    """

    default_parameters = {
        "a_s": 0.3837,
        "b_s": 0.003678,
        "c_s": 0.2,
        "a_f": 3.11,
        "b_f": -0.68,
        "c_f": 2.41,
        "calculate_VW_velocities": False,
        "use_yaw_added_recovery": False,
        "eps_gain": 0.2,
    }

    def __init__(self, parameter_dictionary):
        """
        Stores model parameters for use by methods.

        Args:
            parameter_dictionary (dict): Model-specific parameters.
                Default values are used when a parameter is not included
                in `parameter_dictionary`. Possible key-value pairs include:

                    -   **a_s**: Parameter used to determine the linear
                        relationship between the turbulence intensity and the
                        width of the Gaussian wake shape.
                    -   **b_s**: Parameter used to determine the linear
                        relationship between the turbulence intensity and the
                        width of the Gaussian wake shape.
                    -   **c_s**: Parameter used to determine the linear
                        relationship between the turbulence intensity and the
                        width of the Gaussian wake shape.
                    -   **a_f**: Parameter used to determine super-Gaussian
                        order.
                    -   **b_f**: Parameter used to determine super-Gaussian
                        order.
                    -   **c_f**: Parameter used to determine super-Gaussian
                        order.
                    -   **calculate_VW_velocities**: Flag to enable the
                        calculation of V- and W-component velocities using
                        methods developed in :cite:`bcv-King2019Controls`.
                    -   **use_yaw_added_recovery**: Flag to use yaw added
                        recovery on the wake velocity using methods developed
                        in :cite:`bcv-King2019Controls`.
                    -   **eps_gain**: Tuning value for calculating the V- and
                        W-component velocities using methods developed in
                        :cite:`bcv-King2019Controls`.
        """
        super().__init__(parameter_dictionary)

        self.model_string = "blondel"
        model_dictionary = self._get_model_dict(__class__.default_parameters)

        # wake expansion parameters
        # Table 2 of reference in docstring
        self.a_s = model_dictionary["a_s"]
        self.b_s = model_dictionary["b_s"]
        self.c_s = model_dictionary["c_s"]

        # fitted parameters for super-Gaussian order n
        # Table 3 of reference in docstring
        self.a_f = model_dictionary["a_f"]
        self.b_f = model_dictionary["b_f"]
        self.c_f = model_dictionary["c_f"]

        self.model_grid_resolution = None

        # GCH Parameters
        self.calculate_VW_velocities = model_dictionary["calculate_VW_velocities"]
        self.use_yaw_added_recovery = model_dictionary["use_yaw_added_recovery"]
        self.eps_gain = model_dictionary["eps_gain"]

    def function(
        self,
        x_locations,
        y_locations,
        z_locations,
        turbine,
        turbine_coord,
        deflection_field,
        flow_field,
    ):
        """
        Using the Blondel super-Gaussian wake model, this method calculates and
        returns the wake velocity deficits, caused by the specified turbine,
        relative to the freestream velocities at the grid of points
        comprising the wind farm flow field.

        Args:
            x_locations (np.array): An array of floats that contains the
                streamwise direction grid coordinates of the flow field
                domain (m).
            y_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the direction normal to
                x and parallel to the ground (m).
            z_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the vertical
                direction (m).
            turbine (:py:obj:`floris.simulation.turbine`): Object that
                represents the turbine creating the wake.
            turbine_coord (:py:obj:`floris.utilities.Vec3`): Object containing
                the coordinate of the turbine creating the wake (m).
            deflection_field (np.array): An array of floats that contains the
                amount of wake deflection in meters in the y direction at each
                grid point of the flow field.
            flow_field (:py:class:`floris.simulation.flow_field`): Object
                containing the flow field information for the wind farm.

        Returns:
            np.array, np.array, np.array:
                Three arrays of floats that contain the wake velocity
                deficit in m/s created by the turbine relative to the freestream
                velocities for the U, V, and W components, aligned with the x, y,
                and z directions, respectively. The three arrays contain the
                velocity deficits at each grid point in the flow field.
        """
        # TODO: implement veer
        # Veer (degrees)
        # veer = flow_field.wind_veer

        # Turbulence intensity for wake width calculation
        TI = turbine.current_turbulence_intensity

        # Turbine parameters
        D = turbine.rotor_diameter
        HH = turbine.hub_height
        yaw = -1 * turbine.yaw_angle  # opposite sign convention in this model
        Ct = turbine.Ct
        U_local = flow_field.u_initial

        # Wake deflection
        delta = deflection_field

        # Calculate mask values to mask upstream wake
        yR = y_locations - turbine_coord.x2
        xR = yR * tand(yaw) + turbine_coord.x1

        # Compute scaled variables (Eq 1, pp 3 of ref. [1] in docstring)
        x_tilde = (x_locations - turbine_coord.x1) / D
        r_tilde = (
            np.sqrt(
                (y_locations - turbine_coord.x2 - delta) ** 2 + (z_locations - HH) ** 2,
                dtype=np.float128,
            )
            / D
        )

        # Calculate Beta (Eq 10, pp 5 of ref. [1] in docstring)
        beta = 0.5 * ((1 + np.sqrt(1 - Ct)) / np.sqrt(1 - Ct))

        # Calculate sigma_tilde (Eq 9, pp 5 of ref. [1] in docstring)
        sigma_tilde = (self.a_s * TI + self.b_s) * x_tilde + self.c_s * np.sqrt(beta)

        # Calculate n (Eq 13, pp 6 of ref. [1] in docstring)
        n = self.a_f * np.exp(self.b_f * x_tilde) + self.c_f

        # Calculate max vel def (Eq 5, pp 4 of ref. [1] in docstring)
        a1 = 2 ** (2 / n - 1)
        a2 = 2 ** (4 / n - 2)
        C = a1 - np.sqrt(
            a2
            - (
                (n * Ct)
                * cosd(yaw)
                / (
                    16.0
                    * gamma(2 / n)
                    * np.sign(sigma_tilde)
                    * (np.abs(sigma_tilde) ** (4 / n))
                )
            )
        )

        # Compute wake velocity (Eq 1, pp 3 of ref. [1] in docstring)
        velDef1 = U_local * C * np.exp((-1 * r_tilde ** n) / (2 * sigma_tilde ** 2))
        velDef1[x_locations < xR] = 0

        return (
            np.sqrt(velDef1 ** 2),
            np.zeros(np.shape(velDef1)),
            np.zeros(np.shape(velDef1)),
        )

    @property
    def a_s(self):
        """
        Constant coefficient used in calculation of wake expansion. See
        Eqn. 9 in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._a_s

    @a_s.setter
    def a_s(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for a_s: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._a_s = value
        if value != __class__.default_parameters["a_s"]:
            self.logger.info(
                (
                    "Current value of a_s, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["a_s"])
            )

    @property
    def b_s(self):
        """
        Constant coefficient used in calculation of wake expansion. See
        Eqn. 9 in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._b_s

    @b_s.setter
    def b_s(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for b_s: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._b_s = value
        if value != __class__.default_parameters["b_s"]:
            self.logger.info(
                (
                    "Current value of b_s, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["b_s"])
            )

    @property
    def c_s(self):
        """
        Linear constant used in calculation of wake expansion. See
        Eqn. 9 in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._c_s

    @c_s.setter
    def c_s(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for c_s: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._c_s = value
        if value != __class__.default_parameters["c_s"]:
            self.logger.info(
                (
                    "Current value of c_s, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["c_s"])
            )

    @property
    def a_f(self):
        """
        Constant exponent coefficient used in calculation of the super-Gaussian
        order. See Eqn. 13 in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._a_f

    @a_f.setter
    def a_f(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for a_f: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._a_f = value
        if value != __class__.default_parameters["a_f"]:
            self.logger.info(
                (
                    "Current value of a_f, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["a_f"])
            )

    @property
    def b_f(self):
        """
        Constant exponent coefficient used in calculation of the super-Gaussian
        order. See Eqn. 13 in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._b_f

    @b_f.setter
    def b_f(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for b_f: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._b_f = value
        if value != __class__.default_parameters["b_f"]:
            self.logger.info(
                (
                    "Current value of b_f, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["b_f"])
            )

    @property
    def c_f(self):
        """
        Linear constant used in calculation of the super-Gaussian order. See
        Eqn. 13 in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._c_f

    @c_f.setter
    def c_f(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for c_f: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._c_f = value
        if value != __class__.default_parameters["c_f"]:
            self.logger.info(
                (
                    "Current value of c_f, {0}, is not equal to tuned "
                    + "value of {1}."
                ).format(value, __class__.default_parameters["c_f"])
            )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import numpy as np

from .base_velocity_deficit import VelocityDeficit


class Jensen(VelocityDeficit):
    """
    The Jensen model computes the wake velocity deficit based on the classic
    Jensen/Park model :cite:`jvm-jensen1983note`.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: jvm-
    """

    default_parameters = {"we": 0.05}

    def __init__(self, parameter_dictionary):
        """
        Stores model parameters for use by methods.

        Args:
            parameter_dictionary (dict): Model-specific parameters.
                Default values are used when a parameter is not included
                in `parameter_dictionary`. Possible key-value pairs include:

                -   **we** (*float*): The linear wake decay constant that
                    defines the cone boundary for the wake as well as the
                    velocity deficit. D/2 +/- we*x is the cone boundary for the
                    wake.
        """
        super().__init__(parameter_dictionary)
        self.model_string = "jensen"
        model_dictionary = self._get_model_dict(__class__.default_parameters)
        self.we = float(model_dictionary["we"])

    def function(
        self,
        x_locations,
        y_locations,
        z_locations,
        turbine,
        turbine_coord,
        deflection_field,
        flow_field,
    ):
        """
        Using the Jensen wake model, this method calculates and returns
        the wake velocity deficits, caused by the specified turbine,
        relative to the freestream velocities at the grid of points
        comprising the wind farm flow field.

        Args:
            x_locations (np.array): An array of floats that contains the
                streamwise direction grid coordinates of the flow field
                domain (m).
            y_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the direction normal to
                x and parallel to the ground (m).
            z_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the vertical
                direction (m).
            turbine (:py:obj:`floris.simulation.turbine`): Object that
                represents the turbine creating the wake.
            turbine_coord (:py:obj:`floris.utilities.Vec3`): Object containing
                the coordinate of the turbine creating the wake (m).
            deflection_field (np.array): An array of floats that contains the
                amount of wake deflection in meters in the y direction at each
                grid point of the flow field.
            flow_field (:py:class:`floris.simulation.flow_field`): Object
                containing the flow field information for the wind farm.

        Returns:
            np.array, np.array, np.array:
                Three arrays of floats that contain the wake velocity
                deficit in m/s created by the turbine relative to the freestream
                velocities for the U, V, and W components, aligned with the
                x, y, and z directions, respectively. The three arrays contain
                the velocity deficits at each grid point in the flow field.
        """

        # define the boundary of the wake model ... y = mx + b
        m = self.we
        x = x_locations - turbine_coord.x1
        b = turbine.rotor_radius

        boundary_line = m * x + b

        y_upper = boundary_line + turbine_coord.x2 + deflection_field
        y_lower = -1 * boundary_line + turbine_coord.x2 + deflection_field

        z_upper = boundary_line + turbine.hub_height
        z_lower = -1 * boundary_line + turbine.hub_height

        # calculate the wake velocity
        c = (
            turbine.rotor_diameter
            / (2 * self.we * (x_locations - turbine_coord.x1) + turbine.rotor_diameter)
        ) ** 2

        # filter points upstream and beyond the upper and
        # lower bounds of the wake
        c[x_locations - turbine_coord.x1 < 0] = 0
        c[y_locations > y_upper] = 0
        c[y_locations < y_lower] = 0
        c[z_locations > z_upper] = 0
        c[z_locations < z_lower] = 0

        return (
            2 * turbine.aI * c * flow_field.u_initial,
            np.zeros(np.shape(flow_field.u_initial)),
            np.zeros(np.shape(flow_field.u_initial)),
        )

    @property
    def we(self):
        """
        The linear wake decay constant that defines the cone boundary for the
        wake as well as the velocity deficit. D/2 +/- we*x is the cone boundary
        for the wake.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._we

    @we.setter
    def we(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for we: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._we = value
        if value != __class__.default_parameters["we"]:
            self.logger.info(
                (
                    "Current value of we, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["we"])
            )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import numpy as np

from ....utilities import cosd, sind, tand
from ..base_velocity_deficit import VelocityDeficit


class GaussianModel(VelocityDeficit):
    """
    This is the super-class for all Gaussian-type wake models. It includes
    implementations of functions that subclasses should use to perform
    Gaussian-related calculations (see :cite:`gmb-King2019Controls`)

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: gmb-
    """

    def __init__(self, parameter_dictionary):
        """
        See super-class for initialization details.

        Args:
            parameter_dictionary (dict): Model-specific parameters.
        """
        super().__init__(parameter_dictionary)

    def yaw_added_turbulence_mixing(
        self, coord, turbine, flow_field, x_locations, y_locations, z_locations
    ):
        """
        This method computes the added turbulence due to yaw misalignment.  Yaw misalignment induces additional
        mixing in the flow and causes the wakes to recover faster.
        Args:
            coord (:py:obj:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object.
            flow_field ([type]): [description]
            x_locations (np.array): Streamwise locations in wake.
            y_locations (np.array): Spanwise locations in wake.
            z_locations (np.array): Vertical locations in wake.
        Returns:
            np.array: turbulence mixing adjustment.
        """
        if self.use_yaw_added_recovery:
            # compute turbulence modification
            V, W = self.calc_VW(
                coord, turbine, flow_field, x_locations, y_locations, z_locations
            )

            # calculate fluctuations
            v_prime = flow_field.v + V
            w_prime = flow_field.w + W

            # get u_prime from current turbulence intensity
            u_prime = turbine.u_prime()

            # compute the new TKE
            idx = np.where(
                (np.abs(x_locations - coord.x1) <= turbine.rotor_diameter / 4)
                & (np.abs(y_locations - coord.x2) < turbine.rotor_diameter)
            )
            TKE = (1 / 2) * (
                u_prime ** 2 + np.mean(v_prime[idx]) ** 2 + np.mean(w_prime[idx]) ** 2
            )

            # convert TKE back to TI
            TI_total = turbine.TKE_to_TI(TKE)

            # convert to turbulence due to mixing
            TI_mixing = np.array(TI_total) - turbine.current_turbulence_intensity
        else:
            TI_mixing = 0.0

        return TI_mixing

    def calculate_VW(
        self, V, W, coord, turbine, flow_field, x_locations, y_locations, z_locations
    ):
        """
        This method calculates the V- and W-component velocities using
        methods developed in [1].
        # TODO add reference to 1
        # TODO is this function needed? It simply calls another function

        Args:
            V (np.array): V-component velocity deficits across the flow field.
            W (np.array): W-component velocity deficits across the flow field.
            coord (:py:obj:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object.
            flow_field ([type]): [description]
            x_locations (np.array): Streamwise locations in wake.
            y_locations (np.array): Spanwise locations in wake.
            z_locations (np.array): Vertical locations in wake.

        Raises:
            ValueError: It appears that 'use_yaw_added_recovery' is set
                to True and 'calculate_VW_velocities' is set to False.
                This configuration is not valid. Please set
                'calculate_VW_velocities' to True if you wish to use
                yaw-added recovery.

        Returns:
            np.array, np.array:

                - V-component velocity deficits across the flow field.
                - W-component velocity deficits across the flow field.
        """
        if self.use_yaw_added_recovery:
            if not self.calculate_VW_velocities:
                err_msg = (
                    "It appears that 'use_yaw_added_recovery' is set "
                    + "to True and 'calculate_VW_velocities' is set to False. "
                    + "This configuration is not valid. Please set "
                    + "'calculate_VW_velocities' to True if you wish to use "
                    + "yaw-added recovery."
                )
                self.logger.error(err_msg, stack_info=True)
                raise ValueError(err_msg)
        if self.calculate_VW_velocities:
            V, W = self.calc_VW(
                coord, turbine, flow_field, x_locations, y_locations, z_locations
            )
        return V, W

    def yaw_added_recovery_correction(
        self, U_local, U, W, x_locations, y_locations, turbine, turbine_coord
    ):
        """
        This method corrects the U-component velocities when yaw added recovery
        is enabled. For more details on how the velocities are changed, see [1].
        # TODO add reference to 1

        Args:
            U_local (np.array): U-component velocities across the flow field.
            U (np.array): U-component velocity deficits across the flow field.
            W (np.array): W-component velocity deficits across the flow field.
            x_locations (np.array): Streamwise locations in wake.
            y_locations (np.array): Spanwise locations in wake.
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object.
            turbine_coord (:py:obj:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.

        Returns:
            np.array: U-component velocity deficits across the flow field.
        """
        # compute the velocity without modification
        U1 = U_local - U

        # set dimensions
        D = turbine.rotor_diameter
        xLocs = x_locations - turbine_coord.x1
        ky = self.ka * turbine.current_turbulence_intensity + self.kb
        U2 = (np.mean(W) * xLocs) / ((ky * xLocs + D / 2))
        U_total = U1 + np.nan_to_num(U2)

        # turn it back into a deficit
        U = U_local - U_total

        # zero out anything before the turbine
        U[x_locations < turbine_coord.x1] = 0

        return U

    def calc_VW(
        self, coord, turbine, flow_field, x_locations, y_locations, z_locations
    ):
        """
        This method calculates the V- and W-component velocities using
        methods developed in [1].
        # TODO add reference to 1

        Args:
            coord (:py:obj:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object.
            flow_field ([type]): [description]
            x_locations (np.array): Streamwise locations in wake.
            y_locations (np.array): Spanwise locations in wake.
            z_locations (np.array): Vertical locations in wake.

        Returns:
            np.array, np.array:

                - V-component velocity deficits across the flow field.
                - W-component velocity deficits across the flow field.
        """
        # turbine parameters
        D = turbine.rotor_diameter
        HH = turbine.hub_height
        yaw = turbine.yaw_angle
        Ct = turbine.Ct
        TSR = turbine.tsr
        aI = turbine.aI

        # flow parameters
        Uinf = np.mean(flow_field.wind_map.grid_wind_speed)

        scale = 1.0
        vel_top = (
            Uinf
            * ((HH + D / 2) / flow_field.specified_wind_height) ** flow_field.wind_shear
        ) / Uinf
        vel_bottom = (
            Uinf
            * ((HH - D / 2) / flow_field.specified_wind_height) ** flow_field.wind_shear
        ) / Uinf
        Gamma_top = (
            scale * (np.pi / 8) * D * vel_top * Uinf * Ct * sind(yaw) * cosd(yaw)
        )
        Gamma_bottom = (
            -scale * (np.pi / 8) * D * vel_bottom * Uinf * Ct * sind(yaw) * cosd(yaw)
        )
        Gamma_wake_rotation = (
            0.25 * 2 * np.pi * D * (aI - aI ** 2) * turbine.average_velocity / TSR
        )

        # compute the spanwise and vertical velocities induced by yaw
        eps = self.eps_gain * D  # Use set value

        # decay the vortices as they move downstream - using mixing length
        lmda = D / 8
        kappa = 0.41
        lm = kappa * z_locations / (1 + kappa * z_locations / lmda)
        z = np.linspace(
            z_locations.min(), z_locations.max(), flow_field.u_initial.shape[2]
        )
        dudz_initial = np.gradient(flow_field.u_initial, z, axis=2)
        nu = lm ** 2 * np.abs(dudz_initial[0, :, :])

        # top vortex
        yLocs = y_locations + 0.01 - (coord.x2)
        zT = z_locations + 0.01 - (HH + D / 2)
        rT = yLocs ** 2 + zT ** 2
        V1 = (
            (zT * Gamma_top)
            / (2 * np.pi * rT)
            * (1 - np.exp(-rT / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        W1 = (
            (-yLocs * Gamma_top)
            / (2 * np.pi * rT)
            * (1 - np.exp(-rT / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        # bottom vortex
        zB = z_locations + 0.01 - (HH - D / 2)
        rB = yLocs ** 2 + zB ** 2
        V2 = (
            (zB * Gamma_bottom)
            / (2 * np.pi * rB)
            * (1 - np.exp(-rB / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        W2 = (
            ((-yLocs * Gamma_bottom) / (2 * np.pi * rB))
            * (1 - np.exp(-rB / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        # top vortex - ground
        yLocs = y_locations + 0.01 - (coord.x2)
        zLocs = z_locations + 0.01 + (HH + D / 2)
        V3 = (
            (
                ((zLocs * -Gamma_top) / (2 * np.pi * (yLocs ** 2 + zLocs ** 2)))
                * (1 - np.exp(-(yLocs ** 2 + zLocs ** 2) / (eps ** 2)))
                + 0.0
            )
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        W3 = (
            ((-yLocs * -Gamma_top) / (2 * np.pi * (yLocs ** 2 + zLocs ** 2)))
            * (1 - np.exp(-(yLocs ** 2 + zLocs ** 2) / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        # bottom vortex - ground
        yLocs = y_locations + 0.01 - (coord.x2)
        zLocs = z_locations + 0.01 + (HH - D / 2)
        V4 = (
            (
                ((zLocs * -Gamma_bottom) / (2 * np.pi * (yLocs ** 2 + zLocs ** 2)))
                * (1 - np.exp(-(yLocs ** 2 + zLocs ** 2) / (eps ** 2)))
                + 0.0
            )
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        W4 = (
            ((-yLocs * -Gamma_bottom) / (2 * np.pi * (yLocs ** 2 + zLocs ** 2)))
            * (1 - np.exp(-(yLocs ** 2 + zLocs ** 2) / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        # wake rotation vortex
        zC = z_locations + 0.01 - (HH)
        rC = yLocs ** 2 + zC ** 2
        V5 = (
            (zC * Gamma_wake_rotation)
            / (2 * np.pi * rC)
            * (1 - np.exp(-rC / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        W5 = (
            (-yLocs * Gamma_wake_rotation)
            / (2 * np.pi * rC)
            * (1 - np.exp(-rC / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        # wake rotation vortex - ground effect
        yLocs = y_locations + 0.01 - coord.x2
        zLocs = z_locations + 0.01 + HH
        V6 = (
            (
                (
                    (zLocs * -Gamma_wake_rotation)
                    / (2 * np.pi * (yLocs ** 2 + zLocs ** 2))
                )
                * (1 - np.exp(-(yLocs ** 2 + zLocs ** 2) / (eps ** 2)))
                + 0.0
            )
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        W6 = (
            ((-yLocs * -Gamma_wake_rotation) / (2 * np.pi * (yLocs ** 2 + zLocs ** 2)))
            * (1 - np.exp(-(yLocs ** 2 + zLocs ** 2) / (eps ** 2)))
            * eps ** 2
            / (4 * nu * (x_locations - coord.x1) / Uinf + eps ** 2)
        )

        # total spanwise velocity
        V = V1 + V2 + V3 + V4 + V5 + V6
        W = W1 + W2 + W3 + W4 + W5 + W6

        # no spanwise and vertical velocity upstream of the turbine
        V[
            x_locations < coord.x1 - 1
        ] = 0.0  # Subtract by 1 to avoid numerical issues on rotation
        W[
            x_locations < coord.x1 - 1
        ] = 0.0  # Subtract by 1 to avoid numerical issues on rotation

        W[W < 0] = 0

        return V, W

    @property
    def calculate_VW_velocities(self):
        """
        Flag to enable the calculation of V- and W-component velocities using
        methods developed in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (bool): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._calculate_VW_velocities

    @calculate_VW_velocities.setter
    def calculate_VW_velocities(self, value):
        if type(value) is not bool:
            err_msg = (
                "Value of calculate_VW_velocities must be type "
                + "float; {} given.".format(type(value))
            )
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._calculate_VW_velocities = value

    @property
    def use_yaw_added_recovery(self):
        """
        Flag to use yaw added recovery on the wake velocity using methods
        developed in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (bool): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._use_yaw_added_recovery

    @use_yaw_added_recovery.setter
    def use_yaw_added_recovery(self, value):
        if type(value) is not bool:
            # TODO Shouldn't this be a bool?
            err_msg = (
                "Value of use_yaw_added_recovery must be type "
                + "float; {} given.".format(type(value))
            )
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._use_yaw_added_recovery = value

    @property
    def eps_gain(self):
        """
        Tuning value for calculating the V- and W- component velocities using
        methods developed in [1].

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (bool): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._eps_gain

    @eps_gain.setter
    def eps_gain(self, value):
        if type(value) is not float:
            err_msg = "Value of eps_gain must be type " + "float; {} given.".format(
                type(value)
            )
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._eps_gain = value

    @staticmethod
    def mask_upstream_wake(y_locations, turbine_coord, yaw):
        """
        Calculates values to be used for masking the upstream wake relative to
        the current turbine.

        Args:
            y_locations (np.array): Spanwise locations in wake.
            turbine_coord (:py:obj:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.
            yaw (float): The turbine yaw angle.

        Returns:
            tuple: tuple containing:

                -   yR (np.array): Y locations to mask upstream wake.
                -   xR (np.array): X locations to mask upstream wake.
        """
        yR = y_locations - turbine_coord.x2
        xR = yR * tand(yaw) + turbine_coord.x1
        return xR, yR

    @staticmethod
    def initial_velocity_deficits(U_local, Ct):
        """
        Calculates the initial velocity deficits used in determining the wake
        expansion in a Gaussian wake velocity model.

        Args:
            U_local (np.array): U-component velocities across the flow field.
            Ct (float): The thrust coefficient of a turbine at the current
                operating conditions.

        Returns:
            tuple: tuple containing:

                -   uR (np.array): Initial velocity deficit used in calculation
                    of wake expansion.
                -   u0 (np.array): Initial velocity deficit used in calculation
                    of wake expansion.
        """
        uR = U_local * Ct / (2.0 * (1 - np.sqrt(1 - Ct)))
        u0 = U_local * np.sqrt(1 - Ct)
        return uR, u0

    @staticmethod
    def initial_wake_expansion(turbine, U_local, veer, uR, u0):
        """
        Calculates the initial wake widths associated with wake expansion.

        Args:
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object.
            U_local (np.array): U-component velocities across the flow field.
            veer (float): The amount of veer across the rotor.
            uR (np.array): Initial velocity deficit used in calculation of wake
                expansion.
            u0 (np.array): Initial velocity deficit used in calculation of wake
                expansion.

        Returns:
            tuple: tuple containing:

                -   sigma_y0 (np.array): Initial wake width in the spanwise
                    direction.
                -   sigma_z0 (np.array): Initial wake width in the vertical
                    direction.
        """
        yaw = -1 * turbine.yaw_angle
        sigma_z0 = turbine.rotor_diameter * 0.5 * np.sqrt(uR / (U_local + u0))
        sigma_y0 = sigma_z0 * cosd(yaw) * cosd(veer)
        return sigma_y0, sigma_z0

    @staticmethod
    def gaussian_function(U, C, r, n, sigma):
        """
        A general form of the Gaussian function used in the Gaussian wake
        models.

        Args:
            U (np.array): U-component velocities across the flow field.
            C (np.array): Velocity deficit at the wake center normalized by the
                incoming wake velocity.
            r (float): Radial distance from the wake center.
            n (float): Exponent of radial distance from the wake center.
            sigma (np.array): Standard deviation of the wake.

        Returns:
            np.array: U (np.array): U-component velocity deficits across the
            flow field.
        """
        return U * C * np.exp(-1 * r ** n / (2 * sigma ** 2))



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import json

import numpy as np


class InputReader:
    """
    InputReader parses JSON input files into inputs for the
    :py:class:`~.floris.Floris` class. It handles input validation regarding
    input type, but does not enforce value checking.
    """

    def __init__(self):
        """
        Initializes the parameter types expected in the input data.
        """

        self._valid_objects = ["turbine", "wake", "farm"]

        self._turbine_properties = {
            "rotor_diameter": float,
            "hub_height": float,
            "blade_count": int,
            "pP": float,
            "pT": float,
            "generator_efficiency": float,
            "power_thrust_table": dict,
            "yaw_angle": float,
            "tilt_angle": float,
            "TSR": float,
        }

        self._wake_properties = {
            "velocity_model": str,
            "turbulence_model": str,
            "deflection_model": str,
            "combination_model": str,
            "parameters": dict,
        }

        self._farm_properties = {
            "wind_speed": list,
            "wind_direction": list,
            "turbulence_intensity": list,
            "wind_shear": float,
            "wind_veer": float,
            "air_density": float,
            "layout_x": list,
            "layout_y": list,
            "wind_x": list,
            "wind_y": list,
        }

    def _parseJSON(self, filename):
        """
        Opens the input JSON file and parses the contents into a Python
        dict.

        Args:
            filename (str): Path to the JSON input file.

        Returns:
            dict: The data parsed from the input file.
        """
        with open(filename) as jsonfile:
            data = json.load(jsonfile)
        return data

    def _validate_dict(self, input_dict, type_map):
        """
        Verifies that the expected fields exist in the JSON input file
        and validates the type of the input data by casting the fields
        to appropriate values based on the predefined type maps.

        Args:
            input_dict (dict): Input data.
            type_map (dict): Predefined type-map for type checking inputs;
                structured as {"property": type}.

        Raises:
            KeyError: Missing/invalid key.
            ValueError: Invalid value type.

        Returns:
            dict: Validated and correctly typed input data.
        """
        validated = {}

        # validate the object type
        if "type" not in input_dict:
            raise KeyError("'type' key is required in input")
        if input_dict["type"] not in self._valid_objects:
            raise ValueError(
                "'type' must be one of {}".format(", ".join(self._valid_objects))
            )
        validated["type"] = input_dict["type"]

        # validate the name
        if "name" not in input_dict:
            raise KeyError("'name' key is required in input")
        validated["name"] = input_dict["name"]

        # validate the properties dictionary
        if "properties" not in input_dict:
            raise KeyError("'properties' key is required in input")

        # check every attribute in the predefined type dictionary for existence
        # and proper type in the given inputs
        prop_dict = {}
        properties = input_dict["properties"]
        for element in type_map:
            if element not in properties:
                raise KeyError(
                    "'{}' is required for object type '{}'".format(
                        element, validated["type"]
                    )
                )
            prop_dict[element] = self._cast(type_map[element], properties[element])
        validated["properties"] = prop_dict

        for element in properties:
            if element not in type_map:
                raise KeyError(
                    "'{}' is given but not required for object type '{}'".format(
                        element, validated["type"]
                    )
                )

        return validated

    def _cast(self, typecast, value):
        """
        Casts the input to the given type in `typecast`.

        Args:
            typecast (type): Type class to use for casting.
            value (str): Input to cast to 'typecast'.

        Returns:
            typecast
        """
        return typecast(value)

    def validate_turbine(self, input_dict):
        """
        Checks for the required values and types of input in the given input
        dictionary as required by the :py:obj:`~.turbine.Turbine` object.

        Args:
            input_dict (dict): Input dictionary describing a turbine model.

        Returns:
            dict: A validated dictionary.
        """
        return self._validate_dict(input_dict, self._turbine_properties)

    def validate_wake(self, input_dict):
        """
        Checks for the required values and types of input in the given input
        dictionary as required by the :py:obj:`~.wake.Wake` object.

        Args:
            input_dict (dict): Input dictionary describing a wake model.

        Returns:
            dict: A validated dictionary.
        """
        return self._validate_dict(input_dict, self._wake_properties)

    def validate_farm(self, input_dict):
        """
        Checks for the required values and types of input in the given input
        dictionary as required by the :py:obj:`~.farm.Farm` object.

        Args:
            input_dict (dict): Input dictionary describing a farm model.

        Returns:
            dict: A validated dictionary.
        """
        return self._validate_dict(input_dict, self._farm_properties)

    def read(self, input_file=None, input_dict=None):
        """
        Parses a given input file or input dictionary and validates the
        contents.

        Args:
            input_file (str, optional): A string path to the JSON input file.
                Defaults to None.
            input_dict (dict, optional): A Python dictionary of inputs.
                Defaults to None.

        Raises:
            ValueError: Input file or dictionary must be provided.

        Returns:
            dict, dict, dict, dict:

                - Meta data from the input dictionary.
                - Validated "turbine" section from the input dictionary.
                - Validated "wake" section from the input dictionary.
                - Validated "farm" section from the input dictionary.
        """
        if input_file is not None:
            input_dict = self._parseJSON(input_file)
        elif input_dict is not None:
            input_dict = input_dict.copy()
        else:
            raise ValueError("Input file or dictionary must be provided")

        turbine_dict = input_dict.pop("turbine")
        wake_dict = input_dict.pop("wake")
        farm_dict = input_dict.pop("farm")
        meta_dict = input_dict
        return meta_dict, turbine_dict, wake_dict, farm_dict



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import numpy as np

from ...utilities import cosd, sind
from .base_velocity_deflection import VelocityDeflection


class Jimenez(VelocityDeflection):
    """
    Jiménez wake deflection model, dervied from
    :cite:`jdm-jimenez2010application`.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: jdm-
    """

    default_parameters = {"kd": 0.05, "ad": 0.0, "bd": 0.0}

    def __init__(self, parameter_dictionary):
        """
        Stores model parameters for use by methods.

        Args:
            parameter_dictionary (dict): Model-specific parameters.
                Default values are used when a parameter is not included
                in `parameter_dictionary`. Possible key-value pairs include:

                    -   **kd** (*float*): Parameter used to determine the skew
                        angle of the wake.
                    -   **ad** (*float*): Additional tuning parameter to modify
                        the wake deflection with a lateral offset.
                        Defaults to 0.
                    -   **bd** (*float*): Additional tuning parameter to modify
                        the wake deflection with a lateral offset.
                        Defaults to 0.

        """
        super().__init__(parameter_dictionary)
        self.model_string = "jimenez"
        model_dictionary = self._get_model_dict(__class__.default_parameters)
        self.ad = float(model_dictionary["ad"])
        self.kd = float(model_dictionary["kd"])
        self.bd = float(model_dictionary["bd"])

    def function(
        self, x_locations, y_locations, z_locations, turbine, coord, flow_field
    ):
        """
        Calcualtes the deflection field of the wake in relation to the yaw of
        the turbine. This is coded as defined in [1].

        Args:
            x_locations (np.array): streamwise locations in wake
            y_locations (np.array): spanwise locations in wake
            z_locations (np.array): vertical locations in wake
                (not used in Jiménez)
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object
            coord
                (:py:meth:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.
            flow_field
                (:py:class:`floris.simulation.flow_field.FlowField`):
                Flow field object.

        Returns:
            deflection (np.array): Deflected wake centerline.
        """

        # angle of deflection
        xi_init = cosd(turbine.yaw_angle) * sind(turbine.yaw_angle) * turbine.Ct / 2.0

        x_locations = x_locations - coord.x1

        # yaw displacement
        yYaw_init = (
            xi_init
            * (
                15 * (2 * self.kd * x_locations / turbine.rotor_diameter + 1) ** 4.0
                + xi_init ** 2.0
            )
            / (
                (30 * self.kd / turbine.rotor_diameter)
                * (2 * self.kd * x_locations / turbine.rotor_diameter + 1) ** 5.0
            )
        ) - (xi_init * turbine.rotor_diameter * (15 + xi_init ** 2.0) / (30 * self.kd))

        # corrected yaw displacement with lateral offset
        deflection = yYaw_init + self.ad + self.bd * x_locations

        x = np.unique(x_locations)
        for i in range(len(x)):
            tmp = np.max(deflection[x_locations == x[i]])
            deflection[x_locations == x[i]] = tmp

        return deflection

    @property
    def kd(self):
        """
        Parameter used to determine the skew angle of the wake.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._kd

    @kd.setter
    def kd(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for kd: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._kd = value
        if value != __class__.default_parameters["kd"]:
            self.logger.info(
                (
                    "Current value of kd, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["kd"])
            )

    @property
    def ad(self):
        """
        Parameter available for additional tuning of the wake deflection with a
        lateral offset.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._ad

    @ad.setter
    def ad(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for ad: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._ad = value
        if value != __class__.default_parameters["ad"]:
            self.logger.info(
                (
                    "Current value of ad, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["ad"])
            )

    @property
    def bd(self):
        """
        Parameter available for additional tuning of the wake deflection with a
        lateral offset.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): Value to set.

        Returns:
            float: Value currently set.

        Raises:
            ValueError: Invalid value.
        """
        return self._bd

    @bd.setter
    def bd(self, value):
        if type(value) is not float:
            err_msg = (
                "Invalid value type given for bd: {}, " + "expected float."
            ).format(value)
            self.logger.error(err_msg, stack_info=True)
            raise ValueError(err_msg)
        self._bd = value
        if value != __class__.default_parameters["bd"]:
            self.logger.info(
                (
                    "Current value of bd, {0}, is not equal to tuned " + "value of {1}."
                ).format(value, __class__.default_parameters["bd"])
            )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import scipy as sp

from ..utilities import Vec3


class WindMap:
    """
    Container object to describe initial state of the wind farm atmosphere.
    """

    def __init__(
        self,
        wind_layout,
        layout_array,
        turbulence_intensity=None,
        wind_direction=None,
        wind_speed=None,
    ):
        """
        Computes initial states for the wind direction, wind speed, and
        turbulence intensity, if they are not given directly.

        Args:
            wind_layout ( tuple( list(float), list(float) ) ): The x- and
                y-coordinates of the atmospheric measurement location. This
                argument has the shape:

                    - 0th element: list( x-coordinates)
                    - 1th element: list( y-coordinates)

            layout_array ( list( list(float), list(float) ) ): The x and y
                coordinates of each turbine. This argument has the shape:

                    - 0th element: list( x-coordinates)
                    - 1th element: list( y-coordinates)

            turbulence_intensity (list(float), optional): Turbulence intensity
                values at each measurement location (expressed as a decimal
                fraction). Defaults to None.
            wind_direction (list(float), optional): The wind directions
                (in degrees) at each measurement location. Defaults to None.
            wind_speed (list(float), optional): The wind speed values (in m/s)
                at each measurement location. Defaults to None.
        """
        self.wind_layout = wind_layout
        self.layout_array = layout_array
        self.input_direction = wind_direction
        self.input_speed = wind_speed
        self.input_ti = turbulence_intensity

        # Indicates that fix_wind_layout has been used when True. Initializing to False.
        self.duplicated_wind_layout = False

        if wind_direction is not None:
            self.calculate_wind_direction()
        if wind_speed is not None:
            self.calculate_wind_speed()
        if turbulence_intensity is not None:
            self.calculate_turbulence_intensity()

    # Public functions

    def calculate_wind_speed(self, grid=None, interpolate=False):
        """
        This method calculates the wind speeds at each point,
        interpolated and extrapolated from the input measurements.
        # TODO: describe in detail how this is done.

        Args:
            grid (bool, optional): Switch to enable calculating values
                at grid point coordinates instead of turbine coordinates.
                Defaults to None.
            interpolate (bool, optional): Switch for enabling interpolation
                between given values. Defaults to *False*.
        """
        speed = self.input_speed
        if not grid:
            if all(elem == speed[0] for elem in speed):
                self._turbine_wind_speed = [speed[0]] * len(self.layout_array[0])
            else:
                interpolate = True

        else:
            if all(elem == speed[0] for elem in speed):
                self._grid_wind_speed = np.full(np.shape(self.grid_layout[0]), speed[0])
            else:
                interpolate = True

        if interpolate:
            if self.duplicated_wind_layout and len(self.input_speed) == len(
                self.wind_layout[0]
            ) - len(self.input_speed):
                self._input_speed = self.input_speed + self.input_speed

            if grid is None:
                layout_array = np.array(self.layout_array)
                xp, yp = layout_array[0], layout_array[1]
                newpts = list(zip(xp, yp))
            else:
                layout_array = self.grid_layout
                xp, yp = layout_array[0], layout_array[1]
                newpts = layout_array
            x, y = self.wind_layout[0], self.wind_layout[1]
            z = np.array(speed)
            pts = list(zip(x, y))
            try:
                interp = sp.interpolate.LinearNDInterpolator(pts, z, fill_value=np.nan)
            except Exception:
                self.fix_wind_layout(recalculate="speed", grid=grid)
            else:
                zz = interp(newpts)
                idx = np.where(~np.isnan(zz))
                if np.shape(idx) == (xp.ndim, 0):
                    near = sp.interpolate.NearestNDInterpolator(pts, z)
                else:
                    nearpts = list(zip(xp[idx], yp[idx]))
                    near = sp.interpolate.NearestNDInterpolator(nearpts, zz[idx])
                wind_sp = interp(newpts)
                nearspeed = near(newpts)
                idx = np.where(np.isnan(wind_sp))
                wind_sp[idx] = nearspeed[idx]
                if grid is None:
                    self._turbine_wind_speed = wind_sp.tolist()
                else:
                    self._grid_wind_speed = wind_sp

    def calculate_wind_direction(self, grid=None, interpolate=False):
        """
        This method calculates the wind direction at each point,
        interpolated and extrapolated from the input measurements.
        # TODO: describe in detail how this is done.

        Args:
            grid (bool, optional): Switch to enable calculating values
                at grid point coordinates instead of turbine coordinates.
                Defaults to None.
            interpolate (bool, optional): Switch for enabling interpolation
                between given values. Defaults to *False*.
        """
        wdir = self.input_direction
        if not grid:
            if all(elem == wdir[0] for elem in wdir):
                self._turbine_wind_direction = list(
                    (np.array([wdir[0] - 270] * len(self.layout_array[0])) % 360 + 360)
                    % 360
                )
            else:
                interpolate = True
        else:
            if all(elem == wdir[0] for elem in wdir):
                self._grid_wind_direction = (
                    np.full(np.shape(self.grid_layout[0]), wdir[0] - 270) % 360 + 360
                ) % 360
            else:
                interpolate = True

        if interpolate:
            if self.duplicated_wind_layout and len(self.input_direction) == len(
                self.wind_layout[0]
            ) - len(self.input_direction):
                self._input_direction = self.input_direction + self.input_direction

            if not grid:
                layout_array = np.array(self.layout_array)
                xp, yp = layout_array[0], layout_array[1]
                newpts = list(zip(xp, yp))
            else:
                xt, yt = np.array(self.layout_array)[0], np.array(self.layout_array)[1]
                new_turb_pts = list(zip(xt, yt))
                layout_array = self.grid_layout
                xp, yp = layout_array[0], layout_array[1]
                newpts = layout_array
                turb_wd = np.zeros(np.shape(np.array(self.layout_array)))
            wind_layout_array = np.array(self.wind_layout)
            x, y = wind_layout_array[0], wind_layout_array[1]
            pts = list(zip(x, y))
            a = np.array(self.input_direction)
            wdxy = [np.sin(a * np.pi / 180), np.cos(a * np.pi / 180)]
            wd = np.zeros(np.shape(layout_array))
            for i in range(2):
                z = wdxy[i]
                try:
                    interp = sp.interpolate.LinearNDInterpolator(
                        pts, z, fill_value=np.nan
                    )
                except Exception:
                    self.fix_wind_layout(recalculate="direction", grid=grid)
                else:
                    zz = interp(newpts)
                    idx = np.where(~np.isnan(zz))
                    if np.shape(idx) == (xp.ndim, 0):
                        near = sp.interpolate.NearestNDInterpolator(pts, z)
                    else:
                        nearpts = list(zip(xp[idx], yp[idx]))
                        near = sp.interpolate.NearestNDInterpolator(nearpts, zz[idx])
                    if grid:
                        turb_wd[i] = near(new_turb_pts)
                    wind_dir = interp(newpts)
                    neardir = near(newpts)
                    idx = np.where(np.isnan(wind_dir))
                    wind_dir[idx] = neardir[idx]
                    wd[i] = wind_dir
            widi = (np.arctan2(wd[0], wd[1]) * 180 / np.pi) - 270 % 360
            widi = (widi + 360) % 360
            if not grid:
                self._turbine_wind_direction = widi.tolist()
            else:
                twidi = (np.arctan2(turb_wd[0], turb_wd[1]) * 180 / np.pi) - 270 % 360
                twidi = (twidi + 360) % 360

                self._turbine_wind_direction = twidi.tolist()
                self._grid_wind_direction = widi

    def calculate_turbulence_intensity(self, grid=None, interpolate=False):
        """
        This method calculates the turbulence intensity at each point,
        interpolated and extrapolated from the input measurements.
        # TODO: describe in detail how this is done.

        Args:
            grid (bool, optional): Switch to enable calculating values
                at grid point coordinates instead of turbine coordinates.
                Defaults to None.
            interpolate (bool, optional): Switch for enabling interpolation
                between given values. Defaults to *False*.
        """
        # TODO Add comments!
        ti = self.input_ti
        if not grid:
            if all(elem == ti[0] for elem in ti):
                self._turbine_turbulence_intensity = [ti[0]] * len(self.layout_array[0])
            else:
                interpolate = True
        else:
            ti = self.input_ti
            if all(elem == ti[0] for elem in ti):
                self._grid_turbulence_intensity = np.full(
                    np.shape(self.grid_layout[0]), ti[0]
                )
            else:
                interpolate = True

        if interpolate:
            if self.duplicated_wind_layout and len(self.input_ti) == len(
                self.wind_layout[0]
            ) - len(self.input_ti):
                self._input_ti = self.input_ti + self.input_ti

            if grid is None:
                layout_array = np.array(self.layout_array)
                xp, yp = layout_array[0], layout_array[1]
                newpts = list(zip(xp, yp))
            else:
                layout_array = self.grid_layout
                xp, yp = layout_array[0], layout_array[1]
                newpts = layout_array
            wind_layout_array = np.array(self.wind_layout)
            x, y = wind_layout_array[0], wind_layout_array[1]
            z = np.array(ti)
            pts = list(zip(x, y))
            try:
                interp = sp.interpolate.LinearNDInterpolator(pts, z, fill_value=np.nan)
            except Exception:
                self.fix_wind_layout(recalculate="direction", grid=grid)
            else:
                zz = interp(newpts)
                idx = np.where(~np.isnan(zz))
                if np.shape(idx) == (xp.ndim, 0):
                    near = sp.interpolate.NearestNDInterpolator(pts, z)
                else:
                    nearpts = list(zip(xp[idx], yp[idx]))
                    near = sp.interpolate.NearestNDInterpolator(nearpts, zz[idx])
                wind_t = interp(newpts)
                neart = near(newpts)
                idx = np.where(np.isnan(wind_t))
                wind_t[idx] = neart[idx]
                if grid is None:
                    self._turbine_turbulence_intensity = wind_t.tolist()
                else:
                    self._grid_turbulence_intensity = wind_t

    def fix_wind_layout(self, recalculate, grid=None):
        """
        This method analyzes interpolation errors having to do with
        the wind layout coordinates, and provides a solution.

        Args:
            recalculate (str): A string that indicates which
                method was interupted in order to fix the
                wind layout. The indicated method will be restarted
                after the inputs have been fixed, if not a case of
                unequal input lengths.
            grid (bool, optional): When *True*, this indicates that
                the method being interrupted is interpolating
                is calculating values at flow field grid points.
                Defaults to None.

        #TODO: Have to describe this better.

        Raises:
            SystemExit
        """
        x, y = self.wind_layout[0], self.wind_layout[1]
        xp, yp = x, y
        array_length = len(x)

        # check for case of unequal input lengths, stop all calculations if true
        if array_length != len(y):
            raise SystemExit(
                "FLORIS has stopped because the number of wind input coordinates is not\
                \nequal in the x and y directions."
            )

        elif len(self.input_speed) > 1 and len(self.input_speed) != array_length:
            raise SystemExit(
                "FLORIS has stopped because the number of wind input coordinates is not\
            \nequal to the number of wind input measurements."
            )

        elif (
            len(self.input_direction) > 1 and len(self.input_direction) != array_length
        ):
            raise SystemExit(
                "FLORIS has stopped because the number of wind input coordinates is not\
            \nequal to the number of wind input measurements."
            )

        elif len(self.input_ti) > 1 and len(self.input_ti) != array_length:
            raise SystemExit(
                "FLORIS has stopped because the number of wind input coordinates is not\
            \nequal to the number of wind input measurements."
            )

        # assume error is either:
        # QhullError: QH6214 qhull input error: not enough points(2) to construct initial simplex (need 4)
        # or QhullError: QH6154 Qhull precision error: Initial simplex is flat (facet 1 is coplanar with the interior point)
        # both errors are avoided by adding line of inputs parrell to line originally given
        else:
            # find slope of original wind_layout coords
            j, dx, dy = 0, 0, 0
            while j <= array_length and dx == 0 and dy == 0:
                dx, dy = (x[j + 1] - x[j]), (y[j + 1] - y[j])
                j = j + 1

            # add identical inputs in parallel new line by offsetting points perpendicular to original line
            if dx == 0:
                xp = [min(self.layout_array[0]) - 10] * array_length + [
                    max(self.layout_array[0]) + 10
                ] * array_length
                yp = y + y

            elif dy == 0:
                xp = x + x
                yp = [min(self.layout_array[1]) - 10] * array_length + [
                    max(self.layout_array[1]) + 10
                ] * array_length

            else:
                dist = np.sqrt((dx) ** 2 + (dy) ** 2)
                scaling_param = 500 / dist
                # TODO: find best offset distance (determined by # of turbines or size of domain?)
                # problems with LinearNDInterpolator when offset is small and grid res is high (still runs NearestNDInterpolator)
                for i in range(array_length):
                    xp = xp + [x[i] - (dy * scaling_param)]
                    yp = yp + [y[i] + (dx * scaling_param)]

            # reset input parameters
            if not self.duplicated_wind_layout:
                self.wind_layout = (xp, yp)
                if len(self.input_direction) != 1:
                    self.input_direction = self.input_direction + self.input_direction
                if len(self.input_ti) != 1:
                    self.input_ti = self.input_ti + self.input_ti
                if len(self.input_speed) != 1:
                    self.input_speed = self.input_speed + self.input_speed

            self.duplicated_wind_layout = True

            # restart calculation with new inputs
            if recalculate == "turbulence":
                self.calculate_turbulence_intensity(grid=grid)
            elif recalculate == "direction":
                self.calculate_wind_direction(grid=grid)
            elif recalculate == "speed":
                self.calculate_wind_speed(grid=grid)
            else:
                pass

    # Getters & Setters

    # TODO: describe these properties better.

    @property
    def turbine_wind_speed(self):
        """
        This property returns the wind speeds at each turbine.

        **Note:** This is a virtual property used to "get" a value.

        Returns:
            list(float): Wind speeds at each turbine in m/s.
        """
        return self._turbine_wind_speed

    @property
    def grid_wind_speed(self):
        """
        This property returns the wind speeds at each gridpoint.

        **Note:** This is a virtual property used to "get" a value.

        Returns:
            np.array: Wind speeds at each turbine in m/s.
        """
        return self._grid_wind_speed

    @property
    def turbine_wind_direction(self):
        """
        This property returns the wind direction at each turbine.

        **Note:** This is a virtual property used to "get" a value.

        Returns:
            list(float): Wind directions at each turbine in degrees.
        """
        return self._turbine_wind_direction

    @property
    def grid_wind_direction(self):
        """
        This property returns the wind direction at each gridpoint.

        **Note:** This is a virtual property used to "get" a value.

        Returns:
            np.array: Wind directions at each gridpoint in degrees.
        """
        return self._grid_wind_direction

    @property
    def turbine_turbulence_intensity(self):
        """
        This property returns the turbulence intensity at each turbine.

        **Note:** This is a virtual property used to "get" a value.

        Returns:
            list(float): Turbulence intensities at each turbine in decimal fractions.
        """
        return self._turbine_turbulence_intensity

    @property
    def grid_turbulence_intensity(self):
        """
        This property returns the turbulence intensity at each gridpoint.

        **Note:** This is a virtual property used to "get" a value.

        Returns:
            np.array: Turbulence intensity at each gridpoint as a decimal fraction.
        """
        return self._grid_turbulence_intensity

    @property
    def input_direction(self):
        """
        This property stores and returns the wind directions at each
            input measurement location.

        **Note:** This is a virtual property used to "get" a value.

        Returns:
            list(float): Wind directions expressed in degrees.
        """
        return self._input_direction

    @input_direction.setter
    def input_direction(self, value):
        self._input_direction = value

    @property
    def input_speed(self):
        """
        This property stores and returns the wind speeds at each
            input measurement location.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (list(float)): Wind speeds expressed
                in meters per second.

        Returns:
            list(float)): Wind speeds expressed in meters per second.
        """
        return self._input_speed

    @input_speed.setter
    def input_speed(self, value):
        value = [i if i > 0.0 else 0.0001 for i in value]
        self._input_speed = value

    @property
    def input_ti(self):
        """
        This property stores and returns the turbulence intensity
            at each input measurement location.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (list(float)): Turbulence intensities
                expressed in decimal fraction.

        Returns:
            list(float): Turbulence intensities expressed in decimal fraction.
        """
        return self._input_ti

    @input_ti.setter
    def input_ti(self, value):
        self._input_ti = value

    @property
    def grid_layout(self):
        """
        This property stores and returns an array that contains the
            x and y coordinates of each gridpoint in the flow field.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (np.array): The x- and y-coordinates for each
                gridpoint location.

        Returns:
            np.array: The x- and y-coordinates for each
            gridpoint location.
        """
        return self._grid_layout

    @grid_layout.setter
    def grid_layout(self, value):
        self._grid_layout = value

    @property
    def wind_layout(self):
        """
        This property stores and returns a tuple that contains the
            x and y coordinates of each wind measurement location.

        **Note:** This is a virtual property used to "get" or "set" a value.

        Args:
            value (tuple): Coordinates for each wind measurement location.

        Returns:
            tuple: Coordinates for each wind measurement location.
        """
        return self._wind_layout

    @wind_layout.setter
    def wind_layout(self, value):
        self._wind_layout = value



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize

from .optimization import Optimization


class YawOptimization(Optimization):
    """
    YawOptimization is a subclass of :py:class:`floris.tools.optimization.scipy.
    Optimization` that is used to optimize the yaw angles of all turbines in a Floris
    Farm for a single set of inflow conditions using the SciPy optimize package.
    """

    def __init__(
        self,
        fi,
        minimum_yaw_angle=0.0,
        maximum_yaw_angle=25.0,
        x0=None,
        bnds=None,
        opt_method="SLSQP",
        opt_options=None,
        include_unc=False,
        unc_pmfs=None,
        unc_options=None,
        calc_init_power=True,
    ):
        """
        Instantiate YawOptimization object with a FlorisInterface object
        and assign parameter values.

        Args:
            fi (:py:class:`~.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
            minimum_yaw_angle (float, optional): Minimum constraint on yaw
                angle (deg). Defaults to 0.0.
            maximum_yaw_angle (float, optional): Maximum constraint on yaw
                angle (deg). Defaults to 25.0.
            x0 (iterable, optional): The initial yaw conditions (deg). If none
                are specified, they are set to the current yaw angles for
                all turbines. Defaults to None.
            bnds (iterable, optional): Bounds for the yaw angles (tuples of
                min, max values for each turbine (deg)). If none are
                specified, they are set to (minimum_yaw_angle,
                maximum_yaw_angle) for each turbine. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dictionary, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set to
                {'maxiter': 100, 'disp': False, 'iprint': 1, 'ftol': 1e-7,
                'eps': 0.01}. Defaults to None.
            include_unc (bool, optional): Determines whether wind direction or
                yaw uncertainty are included. If True, uncertainty in wind
                direction and/or yaw position is included when determining
                wind farm power. Uncertainty is included by computing the
                mean wind farm power for a distribution of wind direction
                and yaw position deviations from the intended wind direction
                and yaw angles. Defaults to False.
            unc_pmfs (dictionary, optional): A dictionary containing
                probability mass functions describing the distribution of
                wind direction and yaw position deviations when wind direction
                and/or yaw position uncertainty is included in the power
                calculations. Contains the following key-value pairs:

                -   **wd_unc** (*np.array*): The wind direction
                    deviations from the intended wind direction (deg).
                -   **wd_unc_pmf** (*np.array*): The probability
                    of each wind direction deviation in **wd_unc** occuring.
                -   **yaw_unc** (*np.array*): The yaw angle deviations
                    from the intended yaw angles (deg).
                -   **yaw_unc_pmf** (*np.array*): The probability
                    of each yaw angle deviation in **yaw_unc** occuring.

                If none are specified, default PMFs are calculated using
                values provided in **unc_options**. Defaults to None.
            unc_options (dictionary, optional): A dictionary containing values
                used to create normally-distributed, zero-mean probability mass
                functions describing the distribution of wind direction and yaw
                position deviations when wind direction and/or yaw position
                uncertainty is included. This argument is only used when
                **unc_pmfs** is None and contains the following key-value pairs:

                -   **std_wd** (*float*): The standard deviation of
                    the wind direction deviations from the original wind
                    direction (deg).
                -   **std_yaw** (*float*): The standard deviation of
                    the yaw angle deviations from the original yaw angles (deg).
                -   **pmf_res** (*float*): The resolution in degrees
                    of the wind direction and yaw angle PMFs.
                -   **pdf_cutoff** (*float*): The cumulative
                    distribution function value at which the tails of the
                    PMFs are truncated.

                If none are specified, default values of
                {'std_wd': 4.95, 'std_yaw': 1.75, 'pmf_res': 1.0,
                'pdf_cutoff': 0.995} are used. Defaults to None.
            calc_init_power (bool, optional): If True, calculates initial wind
                farm power for each set of wind conditions. Defaults to True.
        """
        super().__init__(fi)

        if opt_options is None:
            self.opt_options = {
                "maxiter": 50,
                "disp": True,
                "iprint": 2,
                "ftol": 1e-12,
                "eps": 0.1,
            }

        self.unc_pmfs = unc_pmfs

        if unc_options is None:
            self.unc_options = {
                "std_wd": 4.95,
                "std_yaw": 1.75,
                "pmf_res": 1.0,
                "pdf_cutoff": 0.995,
            }

        self.reinitialize_opt(
            minimum_yaw_angle=minimum_yaw_angle,
            maximum_yaw_angle=maximum_yaw_angle,
            x0=x0,
            bnds=bnds,
            opt_method=opt_method,
            opt_options=opt_options,
            include_unc=include_unc,
            unc_pmfs=unc_pmfs,
            unc_options=unc_options,
            calc_init_power=calc_init_power,
        )

    # Private methods

    def _yaw_power_opt(self, yaw_angles):
        yaw_angles = self._unnorm(
            np.array(yaw_angles), self.minimum_yaw_angle, self.maximum_yaw_angle
        )
        return (
            -1
            * self.fi.get_farm_power_for_yaw_angle(
                yaw_angles,
                include_unc=self.include_unc,
                unc_pmfs=self.unc_pmfs,
                unc_options=self.unc_options,
            )
            / self.initial_farm_power
        )

    def _optimize(self):
        """
        Find optimum setting of turbine yaw angles for power production
        given fixed atmospheric conditins (wind speed, direction, etc.).

        Returns:
            opt_yaw_angles (np.array): optimal yaw angles of each turbine.
        """

        self.residual_plant = minimize(
            self._yaw_power_opt,
            self.x0_norm,
            method=self.opt_method,
            bounds=self.bnds_norm,
            options=self.opt_options,
        )

        opt_yaw_angles = self._unnorm(
            self.residual_plant.x, self.minimum_yaw_angle, self.maximum_yaw_angle
        )

        return opt_yaw_angles

    def _set_opt_bounds(self, minimum_yaw_angle, maximum_yaw_angle):
        self.bnds = [(minimum_yaw_angle, maximum_yaw_angle) for _ in range(self.nturbs)]

    # Public methods

    def optimize(self, verbose=True):
        """
        This method solves for the optimum turbine yaw angles for power
        production given a fixed set of atmospheric conditions
        (wind speed, direction, etc.).

        Returns:
            np.array: Optimal yaw angles for each turbine (deg).
        """
        if verbose:
            print("=====================================================")
            print("Optimizing wake redirection control...")
            print("Number of parameters to optimize = ", len(self.x0))
            print("=====================================================")

        opt_yaw_angles = self._optimize()

        if verbose and np.sum(opt_yaw_angles) == 0:
            print(
                "No change in controls suggested for this inflow \
                   condition..."
            )

        return opt_yaw_angles

    def reinitialize_opt(
        self,
        minimum_yaw_angle=None,
        maximum_yaw_angle=None,
        x0=None,
        bnds=None,
        opt_method=None,
        opt_options=None,
        include_unc=None,
        unc_pmfs=None,
        unc_options=None,
        calc_init_power=True,
    ):
        """
        This method reinitializes any optimization parameters that are
        specified. Otherwise, the current parameter values are kept.

        Args:
            minimum_yaw_angle (float, optional): Minimum constraint on yaw
                angle (deg). Defaults to None.
            maximum_yaw_angle (float, optional): Maximum constraint on yaw
                angle (deg). Defaults to None.
            x0 (iterable, optional): The initial yaw conditions (deg). If none
                are specified, they are set to the current yaw angles for all
                turbines. Defaults to None.
            bnds (iterable, optional): Bounds for the yaw angles (tuples of
                min, max values for each turbine (deg)). If none are specified,
                they are set to (minimum_yaw_angle, maximum_yaw_angle) for
                each turbine. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to None.
            opt_options (dictionary, optional): Optimization options used by
                scipy.optimize.minize. Defaults to None.
            include_unc (bool, optional): Determines whether wind direction or
                yaw uncertainty are included. If True, uncertainty in wind
                direction and/or yaw position is included when determining
                wind farm power. Uncertainty is included by computing the
                mean wind farm power for a distribution of wind direction
                and yaw position deviations from the intended wind direction
                and yaw angles. Defaults to None.
            unc_pmfs (dictionary, optional): A dictionary containing
                probability mass functions describing the distribution of
                wind direction and yaw position deviations when wind direction
                and/or yaw position uncertainty is included in the power
                calculations. Contains the following key-value pairs:

                -   **wd_unc** (*np.array*): The wind direction
                    deviations from the intended wind direction (deg).
                -   **wd_unc_pmf** (*np.array*): The probability
                    of each wind direction deviation in **wd_unc** occuring.
                -   **yaw_unc** (*np.array*): The yaw angle deviations
                    from the intended yaw angles (deg).
                -   **yaw_unc_pmf** (*np.array*): The probability
                    of each yaw angle deviation in **yaw_unc** occuring.

                If none are specified, default PMFs are calculated using
                values provided in **unc_options**. Defaults to None.
            unc_options (dictionary, optional): A dictionary containing values
                used to create normally-distributed, zero-mean probability mass
                functions describing the distribution of wind direction and yaw
                position deviations when wind direction and/or yaw position
                uncertainty is included. This argument is only used when
                **unc_pmfs** is None and contains the following key-value pairs:

                -   **std_wd** (*float*): The standard deviation of
                    the wind direction deviations from the original wind
                    direction (deg).
                -   **std_yaw** (*float*): The standard deviation of
                    the yaw angle deviations from the original yaw angles (deg).
                -   **pmf_res** (*float*): The resolution in degrees
                    of the wind direction and yaw angle PMFs.
                -   **pdf_cutoff** (*float*): The cumulative
                    distribution function value at which the tails of the
                    PMFs are truncated.

                If none are specified, default values of
                {'std_wd': 4.95, 'std_yaw': 1.75, 'pmf_res': 1.0,
                'pdf_cutoff': 0.995} are used. Defaults to None.
            calc_init_power (bool, optional): If True, calculates initial wind
                farm power for each set of wind conditions. Defaults to True.
        """
        if minimum_yaw_angle is not None:
            self.minimum_yaw_angle = minimum_yaw_angle
        if maximum_yaw_angle is not None:
            self.maximum_yaw_angle = maximum_yaw_angle
        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = [
                turbine.yaw_angle
                for turbine in self.fi.floris.farm.turbine_map.turbines
            ]
        self.x0_norm = self._norm(
            np.array(self.x0), self.minimum_yaw_angle, self.maximum_yaw_angle
        )
        if bnds is not None:
            self.bnds = bnds
            self.minimum_yaw_angle = np.min([bnds[i][0] for i in range(self.nturbs)])
            self.maximum_yaw_angle = np.max([bnds[i][1] for i in range(self.nturbs)])
        else:
            self._set_opt_bounds(self.minimum_yaw_angle, self.maximum_yaw_angle)
        self.bnds_norm = [
            (
                self._norm(
                    self.bnds[i][0], self.minimum_yaw_angle, self.maximum_yaw_angle
                ),
                self._norm(
                    self.bnds[i][1], self.minimum_yaw_angle, self.maximum_yaw_angle
                ),
            )
            for i in range(self.nturbs)
        ]
        if opt_method is not None:
            self.opt_method = opt_method
        if opt_options is not None:
            self.opt_options = opt_options
        if include_unc is not None:
            self.include_unc = include_unc
        if unc_pmfs is not None:
            self.unc_pmfs = unc_pmfs
        if unc_options is not None:
            self.unc_options = unc_options

        if self.include_unc & (self.unc_pmfs is None):
            if self.unc_options is None:
                self.unc_options = {
                    "std_wd": 4.95,
                    "std_yaw": 1.75,
                    "pmf_res": 1.0,
                    "pdf_cutoff": 0.995,
                }

            # create normally distributed wd and yaw uncertainty pmfs
            if self.unc_options["std_wd"] > 0:
                wd_bnd = int(
                    np.ceil(
                        norm.ppf(
                            self.unc_options["pdf_cutoff"],
                            scale=self.unc_options["std_wd"],
                        )
                        / self.unc_options["pmf_res"]
                    )
                )
                wd_unc = np.linspace(
                    -1 * wd_bnd * self.unc_options["pmf_res"],
                    wd_bnd * self.unc_options["pmf_res"],
                    2 * wd_bnd + 1,
                )
                wd_unc_pmf = norm.pdf(wd_unc, scale=self.unc_options["std_wd"])
                # normalize so sum = 1.0
                wd_unc_pmf = wd_unc_pmf / np.sum(wd_unc_pmf)
            else:
                wd_unc = np.zeros(1)
                wd_unc_pmf = np.ones(1)

            if self.unc_options["std_yaw"] > 0:
                yaw_bnd = int(
                    np.ceil(
                        norm.ppf(
                            self.unc_options["pdf_cutoff"],
                            scale=self.unc_options["std_yaw"],
                        )
                        / self.unc_options["pmf_res"]
                    )
                )
                yaw_unc = np.linspace(
                    -1 * yaw_bnd * self.unc_options["pmf_res"],
                    yaw_bnd * self.unc_options["pmf_res"],
                    2 * yaw_bnd + 1,
                )
                yaw_unc_pmf = norm.pdf(yaw_unc, scale=self.unc_options["std_yaw"])
                # normalize so sum = 1.0
                yaw_unc_pmf = yaw_unc_pmf / np.sum(yaw_unc_pmf)
            else:
                yaw_unc = np.zeros(1)
                yaw_unc_pmf = np.ones(1)

            self.unc_pmfs = {
                "wd_unc": wd_unc,
                "wd_unc_pmf": wd_unc_pmf,
                "yaw_unc": yaw_unc,
                "yaw_unc_pmf": yaw_unc_pmf,
            }

        if calc_init_power:
            self.initial_farm_power = self.fi.get_farm_power_for_yaw_angle(
                [0.0] * self.nturbs,
                include_unc=include_unc,
                unc_pmfs=unc_pmfs,
                unc_options=unc_options,
            )

    # Properties

    @property
    def minimum_yaw_angle(self):
        """
        The minimum yaw angle for the optimization. The setting-method
        updates the optimization bounds accordingly.

        **Note**: This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): The minimum yaw angle to set (deg).

        Returns:
            float: The minimum yaw angle currently set (deg).
        """
        return self._minimum_yaw_angle

    @minimum_yaw_angle.setter
    def minimum_yaw_angle(self, value):
        self._minimum_yaw_angle = value

    @property
    def maximum_yaw_angle(self):
        """
        The maximum yaw angle for the optimization. The setting-method
        updates the optimization bounds accordingly.

        **Note**: This is a virtual property used to "get" or "set" a value.

        Args:
            value (float): The maximum yaw angle to set (deg).

        Returns:
            float: The maximum yaw angle currently set (deg).
        """
        return self._maximum_yaw_angle

    @maximum_yaw_angle.setter
    def maximum_yaw_angle(self, value):
        self._maximum_yaw_angle = value

    @property
    def x0(self):
        """
        The initial yaw angles used for the optimization.

        **Note**: This is a virtual property used to "get" or "set" a value.

        Args:
            value (iterable): The yaw angle initial conditions to set (deg).

        Returns:
            list: The yaw angle initial conditions currently set (deg).
        """
        return self._x0

    @x0.setter
    def x0(self, value):
        self._x0 = value
